''************************
''*   MIDI file player   *
''************************
'' Author:        Andy Schenk, Insonix
'' Date:          March 2009
'' Modified by:   James Long
'' Date:          April 2009

'' uses the PhaseMod GM Synthesizer to play the MIDI files of a play list.
'' Needs an SD card connected, and the MIDI files stored on the card.
'' Because the files are played directly from the card, only the MidiFile Format 0
'' is supported (all Tracks mixed down to one).


CON
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000

  DO              = 0                                   '<-- SD card Interface pins
  CLK             = 1
  DI              = 2
  CS              = 3
  
  Left            = 10                                  '<-- Audio Out pins                        
  Right           = 11
  CR              = 13                                  ' ANSII Carriage Return character
  LF              = 10                                  ' ANSII Line Feed Character
  CLS             = 12
  maxScriptSize   = 128                                 ' Maximum number of values in a Config file  
        
VAR

  byte  filename[16], cueList[128]      
  long  time, tq, ts, us, ioControl[2]  

OBJ
  sd            : "fsrwFemto"
  synth         : "pm_synth_20"
  disp          : "Serial_Lcd"
  lightout      : "PCF8575"
  coguse        : "Sparecogs"
        
PUB Main | dispcog

  sd.start(@ioControl)                                  'Start SD card driver
  lightout.start(@cuelist)                              'Start the outputs and turn the outputs off
  disp.init(7,19200,2)                                  'Start the LCD Display
  disp.cursor(0)                                        'Set the cursor to 0 (invisible)  
  disp.cls                                             
  synth.start(Left,Right,2)                             'Start synth with 20 voices
  disp.str(string("Blue Organ"))                        'ShoW the user the Unit name
  disp.putc($0D)                                        'Move the LCD cursor to the next line
  disp.str(string("Version 1.0"))                       'Show the firmware Version
  waitcnt(clkfreq*5 + cnt)
  if \sd.mount(DO, Clk, DI, CS) == 0                    'Mount the SD card
    if \LoadOutputs                                     'Load the Output file
      Outputloadfail                                    ' tell the user if it failed
      repeat                                            ' and hang the system until he fixes it  
    filename := @playlist                               'Start Play List
    repeat
      disp.cls                                          'Clear the Display
      disp.str(string("Free cogs: ",$0D))               'Show the amount of free Cogs
      dispcog := coguse.freestring
      disp.str(dispcog)
      waitcnt(clkfreq*2 + cnt)
      disp.cls                                          'Clear the Display
      disp.str(string("Playing: ",$0D))                 'Show the file to be played
      disp.str(filename)
      waitcnt(clkfreq*2 + cnt)
      midiplay(filename)                                'play the file
      filename += strsize(filename)+1
    until filename => @listend                          'next file
    synth.allOff
    disp.cls
    disp.str(string($0D,"All Songs done"))
  else
    disp.cls
    disp.home
    disp.str(string("SD card not", $0D))
    disp.str(string("found!"))

  repeat

PUB midiplay(name)  | lg, dt, b, v, st, d1, d2, ch, nout
  us := clkfreq / 1_000_000                             'ticks / microsecond
  b := st := d1 := d2 := v := 0
  ifnot sd.popen(name,"r")
    sd.pread(@v,4)
    if v == "M" + "T"<<8 + "h"<<16 + "d"<<24            'Header ID
      sd.pread(@v,4)                                    'Header Len
      v~
      sd.pread(@v,2)                                    'Format
      if v <> 0
        disp.cls
        disp.str(string("Only Format 0", CR, "Supported!"))
        waitcnt(clkfreq*2 + cnt)
      else
        sd.pread(@v,2)                                  'Tracks
        sd.pread(@v,2)                                  'Ticks/Quarter
        tq := v>>8 + (v&255)<<8
        ts := 500_000 / tq                              'default Tempo 120 BPM
        sd.pread(@v,4)                                  '"MTrk"
        sd.pread(@v,4)                                  'Track length
        lg := v>>24 + (v>>16&255)<<8 + (v>>8&255)<<16
        time := cnt
        repeat
          sd.pread(@b,1)                                'get delta time
          lg--
          dt := b & $7F
          repeat while b>127
            sd.pread(@b,1)
            lg--
            dt := dt<<7 + (b & $7F)
          time := time + dt*us*ts
          repeat until time-cnt < 0                     'wait deltatime
          b~
          sd.pread(@b,1)                                'MIDI byte
          if b > 127
            st := b                                     'new status byte
          elseif st==$FF
            st~
          ch := st & $0F
          d1~
          d2~
          case st & $F0                                 'decode MIDI Event
            $90: sd.pread(@d1,1)                        'Note 
                 sd.pread(@d2,1)                        'Velocity
                 lg -= 3
                 if d2>0
                   synth.noteOn(d1,ch,d2)               'Note On  
                   lightout.main($90,d1,@cuelist)              
                   'show(15,ch+1,string(127),-1)
                 else
                   synth.noteOff(d1,ch)                 'Note Off if Vel=0
                   lightout.main($80,d1,@cuelist)
                   'show(15,ch+1,string("."),-1)
            $80: sd.pread(@d1,1)                        'Note Off 
                 sd.pread(@d2,1)                        'Velocity
                 lg -= 3
                 synth.noteOff(d1,ch)
                 lightout.main($80,d1,@cuelist)
                 'show(15,ch+1,string("."),-1)
                 if ch==9
                   'show(20,ch+1,string(">"),d1)
            $A0: sd.pread(@d1,1)                        'Poly AfterTouch 
                 sd.pread(@d2,1)                        '(not supported)
                 lg -= 3
            $C0: sd.pread(@d1,1)                        'Prg Change 
                 lg -= 2
                 synth.prgChange(d1,ch)
                 'show(0,ch+1,string("Ch:"),ch+1)
                 'show(6,ch+1,string("Prg:"),d1)
            $D0: sd.pread(@d1,1)                        'Mono AfterTouch 
                 lg -= 2                                '(not supported)
            $B0: sd.pread(@d1,1)                        'Controller (nr) 
                 sd.pread(@d2,1)                        'value
                 lg -= 3
                 if d1==7
                   synth.volContr(d2,ch)                '7=Vol
                 if d1==10
                   synth.panContr(d2,ch)                '10=Pan
            $E0: sd.pread(@d1,1)                        'Pitch Bender 
                 sd.pread(@d2,1)                        '(not supported!)
                 lg -= 3
            $F0: sd.pread(@d1,1)                        'Meta Event
                 v~
                 if st==$F0 or st==$F7                  'SysEx
                   repeat
                     sd.pread(@v,1)
                     lg--
                     d1--
                   until v==$F7 or b=< 0
                   d1~
                 else
                   lg -= 2
                   sd.pread(@b,1)                       'Len
                   repeat while b>0                     'read meta data
                     sd.pread(@d2,1)
                     v := v<<8 + d2
                     b--
                     lg--
                   if d1==81                            'Tempo
                     ts := v / tq
            0:   sd.pread(@d1,1)                        'Meta Event Running status
                 lg -= 2
                 sd.pread(@b,1)                         'Len
                 repeat while b>0                       'read meta data
                   sd.pread(@d2,1)
                   v := v<<8 + d2
                   b--
                   lg--
            
        until lg < 1
    else
      disp.str(string("Not a", CR, "MIDI file"))
    sd.pclose
    synth.allOff
  else
    disp.cls
    disp.str(string("File not", CR, "found"))
    waitcnt(clkfreq*3 + cnt)

Pub LoadOutputs | _r, _cursor, _cuect, _outnum, _midinote, _out

'' Load the script from the SD card
    _cursor := @cueList                  
    _r := sd.popen(@Outputfile, "r")                    ' and open the script file
    if _r < 0                                           ' Check for successful open
      disp.cls
      disp.str(string("Config file", CR, "fail!"))
      waitcnt(clkfreq*3 + cnt)
      abort                                             ' and abort if it failed
    disp.cls
    disp.str(string("Loading Config!"))
    waitcnt(clkfreq + cnt)
    _cueCt := 0
    repeat 
      _outnum := getnum                                 ' Get the output number
      _midinote := getnum                               ' Get the note from the file       
      bytemove(_cursor,@_midinote,1)                    ' Move the note into the word array 
      _cursor += 1                                      ' Move the cursor to the next values
      _cueCt  += 1                                          
      if _cueCt < maxScriptSize
        next
      elseif _cueCt > maxScriptSize
        bigscript                                       ' the array - if we do, tell the
        disp.cls
        disp.str(string("We had a", CR, "size problem!"))
        waitcnt(clkfreq + cnt)
        abort                                           ' user and abort 
      elseif _cueCt == maxScriptSize                    ' and be sure we don't overflow
        disp.cls
        disp.str(string("Saved all", CR, "Outputs"))
        waitcnt(clkfreq*3 + cnt)
        quit    
    'cuectmsg(_cuect)                                   ' and let them know we got it loaded
    'waitcnt(clkfreq*5 + cnt)  

PRI getnum | n, num

'' Get a number or letter from the file

  num := 0
  repeat
    n := sd.pgetc                                       ' Get a character
    if n < 0                                            ' Return -1 for end of file
      abort
    case n
      13: quit                                          ' LF and
      44: quit                                          ' comma end a number
      48..57: num := num * 10 + n-48                    ' Handle a digit
      65..76: num := n-64                               ' Map letters A thru L to 1 thru 12 
  return num

PUB cuectmsg(ct)| tostr

  disp.cls
  disp.str(@MsgLoaded)
  waitcnt(clkfreq*2 + cnt)
  disp.cls
  disp.str(@ct)
  waitcnt(clkfreq*2 + cnt)
  disp.cls
  disp.str(@MsgLoadedCues)

PUB bigscript

  disp.cls
  disp.str(@MsgTooBig)
  waitcnt(clkfreq*3 + cnt)

PUB Outputloadfail

  disp.cls
  disp.str(@MsgLoadFailed)
  waitcnt(clkfreq*3 + cnt)

DAT

playlist
                        byte    "26er.mid", 0
                        byte    "sundonSD.mid", 0
                        byte    "GlenM.mid", 0
                        byte    "Bei.mid", 0
                        byte    "Piguin.mid", 0
                        byte    "Barb.mid", 0
                        byte    "Lava1.mid", 0
                        byte    "NUTRAC.MID", 0
                        byte    "BOHEMIAN.MID", 0
                        byte    "POPCORN0.MID", 0
listend
  
Outputfile            byte    "config.csv", 0
MsgLoaded             byte    "Config loaded with ", 0
MsgLoadedCues         byte    " outputs", lf, lf, 0
MsgLoadFailed         byte    "Output file", cr, "load failed", lf, 0
MsgTooBig             byte    "Config is too big", lf, 0
 
     
{{
 ---------------------------------------------------------------------------
                Terms of use: MIT License                                   
 ---------------------------------------------------------------------------
   Permission is hereby granted, free of charge, to any person obtaining a  
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation 
  the rights to use, copy, modify, merge, publish, distribute, sublicense,  
    and/or sell copies of the Software, and to permit persons to whom the   
    Software is furnished to do so, subject to the following conditions:    
                                                                            
   The above copyright notice and this permission notice shall be included  
           in all copies or substantial portions of the Software.           
                                                                            
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING   
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER     
                       DEALINGS IN THE SOFTWARE.                            
 ---------------------------------------------------------------------------
}}  